local ElkFactions = LibStub("AceAddon-3.0"):NewAddon("ElkFactions", "AceBucket-3.0", "AceConsole-3.0", "AceHook-3.0")

local string_format = string.format
local string_gsub   = string.gsub

local MAX_FACTION_INDEX = 200

local LDB = LibStub("LibDataBroker-1.1")
local LDBIcon = LibStub("LibDBIcon-1.0")
local LibQTip = LibStub("LibQTip-1.0")

local gender = UnitSex("player")

local reputationSessionBase = {}

local DBdefaults = {
	profile = {
		minimap = {
			hide = false,
		},
	},
	char = {
		collapsedHeaders = {
		},
	},
}


-- -----
-- HELPER FUNCTIONS
-- -----

local watchedFactionIndex = nil
local function GetWatchedFactionIndex()
	local watchedName = GetWatchedFactionInfo()
	if not watchedName then
		-- no faction is currently watched
		watchedFactionIndex = nil
		return nil
	end
	if watchedFactionIndex and GetFactionInfo(watchedFactionIndex) == watchedName then
		-- we have an index cached and it's still correct
		return watchedFactionIndex
	end
	
	local factionIndex = 1
	repeat
		if select(12, GetFactionInfo(factionIndex)) then
			watchedFactionIndex = factionIndex
			return watchedFactionIndex
		end
		factionIndex = factionIndex + 1
	until factionIndex > MAX_FACTION_INDEX
end

local function SetWatchedFactionName(factionName)
--	self:Print("SetWatchedFactionName(", factionname, ") called...")
	if not factionname then
		SetWatchedFactionIndex(0)
		return
	end
	local factionIndex = GetFactionIndexByName(factionName)
	if factionIndex then
		SetWatchedFactionIndex(factionIndex)
	end
end

local function GetFactionIndexByName(factionName)
	local factionIndex = 1
	local lastFactionName
	repeat
		local name = GetFactionInfo(factionIndex)
		if name == lastFactionName then break end	-- same name twice in a row -> invalid data -> we're done
		lastFactionName = name
		if name == factionName then
			return factionIndex
		end
		factionIndex = factionIndex + 1
	until factionIndex > MAX_FACTION_INDEX
end

-- -----
local function do_OnEnter(frame)
	local tooltip = LibQTip:Acquire("ElkFactionsTip")
	tooltip:SmartAnchorTo(frame)
	tooltip:SetAutoHideDelay(0.1, frame)
	tooltip:EnableMouse(true)
	ElkFactions:UpdateTooltip()
	tooltip:Show()
end

local function do_OnLeave()
	-- empty dummy
end

local function do_OnClick(frame, button)
	if button == "RightButton" then
		InterfaceOptionsFrame_OpenToCategory("ElkFactions")
	else
		if IsShiftKeyDown() then
			local watchedIndex = GetWatchedFactionIndex()
			if watchedIndex then
				local activeWindow = ChatEdit_GetActiveWindow()
				if activeWindow then
					local text = ElkFactions:GetFactionText(nil, watchedIndex, false)
					activeWindow:Insert(text)
				end
			end
		else
			ToggleCharacter("ReputationFrame")
		end
	end
end

function ElkFactions:OnInitialize()
	self.dbo = LDB:NewDataObject("ElkFactions", {
			type = "data source",
			text = "",
			icon = [[Interface\Icons\Spell_Holy_SealOfSalvation]],
			OnEnter = do_OnEnter,
			OnLeave = do_OnLeave,
			OnClick = do_OnClick,
		})

	self.db = LibStub("AceDB-3.0"):New("ElkFactionsDB", DBdefaults, true)
	self.db.RegisterCallback(self, "OnProfileChanged", "RefreshConfig")
	self.db.RegisterCallback(self, "OnProfileCopied", "RefreshConfig")
	self.db.RegisterCallback(self, "OnProfileReset", "RefreshConfig")

	LDBIcon:Register("ElkFactions", self.dbo, self.db.profile.minimap)
end

function ElkFactions:OnEnable()
	-- events for faction data changes
	self:RegisterBucketEvent({"UPDATE_FACTION"}, .1, "FactionUpdate")
	
	-- events for zone changes
--~ 	self:RegisterBucketEvent({"ZONE_CHANGED", "ZONE_CHANGED_INDOORS", "ZONE_CHANGED_NEW_AREA", "MINIMAP_ZONE_CHANGED", "PLAYER_ENTERING_WORLD"}, .1, "ZoneUpdate")

	-- events for collapsing headers
	self:RegisterBucketEvent({"PLAYER_ENTERING_WORLD"}, .1, "CollapseHeaders")

	-- hook UI functions to be informed of some user actions
	self:Hook("CollapseFactionHeader", "Hook_CollapseFactionHeader", true)
	self:Hook("ExpandFactionHeader", "Hook_ExpandFactionHeader", true)
	self:Hook("SetWatchedFactionIndex", "Hook_SetWatchedFactionIndex", true)

	-- refresh session's base reputations
	self:UpdateBaseReputations()
end

function ElkFactions:RefreshConfig()
	-- minimap icon
	LDBIcon:Refresh("ElkFactions", self.db.profile.minimap)
	
	-- collapsed headers
	self:CollapseHeaders()
	
	-- auto faction
--~ 	self:ApplyAutoZoneFaction()
	
	-- LDB text
	self:UpdateText()
	
	-- tooltip
	self:UpdateTooltip()
end

function ElkFactions:FactionUpdate()
	-- refresh session's base reputations
	self:UpdateBaseReputations()
	
	-- update the LDB text
	self:UpdateText()
	
	-- update the tooltip
	self:UpdateTooltip()
end

function ElkFactions:UpdateBaseReputations()
	local factionIndex = 1
	local lastFactionName
	repeat
		local name, description, standingID, barMin, barMax, barValue, atWarWith, canToggleAtWar, isHeader, isCollapsed, hasRep, isWatched, isChild = GetFactionInfo(factionIndex)
		if name == lastFactionName then break end	-- same name twice in a row -> invalid data -> we're done
		lastFactionName = name

		if name and not reputationSessionBase[name] then
			reputationSessionBase[name] = barValue
		end

		factionIndex = factionIndex + 1
	until factionIndex > MAX_FACTION_INDEX
end

local currentZoneCached
function ElkFactions:ZoneUpdate()
	local currentZone = GetRealZoneText().." - "..GetSubZoneText()
	-- if the current zone changed, refresh the automatic zone faction
	if currentZone ~= currentZoneCached then
		self:ApplyAutoZoneFaction()
		currentZoneCached = currentZone
	end
end

function ElkFactions:UpdateText()
	local watchedIndex = GetWatchedFactionIndex()
	if watchedIndex then
		self.dbo.text = self:GetFactionText(nil, watchedIndex, true)
	else
		self.dbo.text = ""
	end
end

local FACTION_COLORS_BLIZZARD = setmetatable({}, {__index = function(self, key) local color = FACTION_BAR_COLORS[key]; local value = ("|cff%.2x%.2x%.2x"):format(color.r * 255, color.g * 255, color.b * 255); rawset(self, key, value); return value end })
local FACTION_COLORS_INC_old = {
	[1] = "|cff8c0000",			-- hated
	[2] = "|cffff0000",			-- hostile
	[3] = "|cffff8c00",			-- unfriendly
	[4] = "|cffbfbfbf",			-- neutral
	[5] = "|cffffffff",			-- friendly
	[6] = "|cff00ff00",			-- honored
	[7] = "|cff3f66e5",			-- reverted
	[8] = "|cff9933cc",			-- exalted
}
local FACTION_COLORS_INC = {
	[1] = "|cffff0000",			-- hated
	[2] = "|cffff8c00",			-- hostile
	[3] = "|cffffff00",			-- unfriendly
	[4] = "|cffffffff",			-- neutral
	[5] = "|cff00ff00",			-- friendly
	[6] = "|cff3f66e5",			-- honored
	[7] = "|cff9933cc",			-- reverted
	[8] = "|cffe6cc80",			-- exalted
}

local function GetFactionDeltaText(factionDelta)
	return (factionDelta == 0  and "-") or (factionDelta > 0 and "+"..factionDelta) or factionDelta
end

local function GetFactionDeltaColor(factionDelta)
	return (factionDelta == 0 and "|cffffffff") or (factionDelta > 0 and "|cff00ff00") or ("|cffff0000")
end

local function tooltip_line_OnMouseUp(frame, factionIndex, button)
	local name, description, standingID, barMin, barMax, barValue, atWarWith, canToggleAtWar, isHeader, isCollapsed, hasRep, isWatched, isChild = GetFactionInfo(factionIndex)
	if IsControlKeyDown() then
		if isHeader then
			if isCollapsed then
				ExpandFactionHeader(factionIndex)
			else
				CollapseFactionHeader(factionIndex)
			end
		else
			if IsFactionInactive(factionIndex) then
				SetFactionActive(factionIndex)
			else
				SetFactionInactive(factionIndex)
			end
		end
	elseif IsShiftKeyDown() then
		local activeWindow = ChatEdit_GetActiveWindow()
		if activeWindow then
			local repIs = barValue - barMin
			local repMax = barMax - barMin
			local text = ElkFactions:GetFactionText(nil, factionIndex, false)
			activeWindow:Insert(text)
		end
		return
--~ 	elseif IsAltKeyDown() then
--~ 		local curZone = GetRealZoneText()
--~ 		if self.db.profile.AutoZoneFaction[curZone] and self.db.profile.AutoZoneFaction[curZone] == name then
--~ 			self.db.profile.AutoZoneFaction[curZone] = nil
--~ 		else
--~ 			self.db.profile.AutoZoneFaction[curZone] = name
--~ 		end
--~ 		self:SetAutoFaction(curZone)
	elseif not isHeader or hasRep then
		if isWatched then
			PlaySound("igMainMenuOptionCheckBoxOff")
			SetWatchedFactionIndex(0)
		else
			PlaySound("igMainMenuOptionCheckBoxOn")
			SetWatchedFactionIndex(factionIndex)
		end
	end
end

function ElkFactions:UpdateTooltip()
	if not LibQTip:IsAcquired("ElkFactionsTip") then
		return
	end
	local tooltip = LibQTip:Acquire("ElkFactionsTip")
	tooltip:Clear()
	tooltip:SetColumnLayout(5, "LEFT", "CENTER", "CENTER", "RIGHT", "RIGHT")

	local curLine

	curLine = tooltip:AddHeader()
	tooltip:SetCell(curLine, 1, "|cfffed100Elk|cffffffffFactions", tooltip:GetHeaderFont(), "CENTER", 5)

-- -----
--~ 	local colortest_inc = "  "
--~ 	local colortest_inc_old = "  "
--~ 	local colortest_blizz = "  "
--~ 	for i = 1, 8 do
--~ 		colortest_inc		= colortest_inc		..FACTION_COLORS_INC[i]			..GetText("FACTION_STANDING_LABEL"..i, gender).."  "
--~ 		colortest_inc_old	= colortest_inc_old	..FACTION_COLORS_INC_old[i]		..GetText("FACTION_STANDING_LABEL"..i, gender).."  "
--~ 		colortest_blizz		= colortest_blizz	..FACTION_COLORS_BLIZZARD[i]	..GetText("FACTION_STANDING_LABEL"..i, gender).."  "
--~ 	end
--~ 	curLine = tooltip:AddLine()
--~ 	tooltip:SetCell(curLine, 1, colortest_inc, nil, "CENTER", 5)
--~ 	curLine = tooltip:AddLine()
--~ 	tooltip:SetCell(curLine, 1, colortest_inc_old, nil, "CENTER", 5)
--~ 	curLine = tooltip:AddLine()
--~ 	tooltip:SetCell(curLine, 1, colortest_blizz, nil, "CENTER", 5)
--~ 	curLine = tooltip:AddLine(" ")
-- -----

	local numFactions = GetNumFactions()
	local header_main = nil
	local header_child = nil
	local indentation
	local name, description, standingID, barMin, barMax, barValue, atWarWith, canToggleAtWar, isHeader, isCollapsed, hasRep, isWatched, isChild
	for factionIndex=1, numFactions do
		name, description, standingID, barMin, barMax, barValue, atWarWith, canToggleAtWar, isHeader, isCollapsed, hasRep, isWatched, isChild = GetFactionInfo(factionIndex)
		if isHeader then
			if isChild then
				header_child = name
				indentation = isCollapsed and "      |TInterface\\Buttons\\UI-PlusButton-Up:20:20|t" or "      |TInterface\\Buttons\\UI-MinusButton-Up:20:20|t"
			else
				header_child = nil
				header_main = name
				indentation = isCollapsed and "|TInterface\\Buttons\\UI-PlusButton-Up:20:20|t" or "|TInterface\\Buttons\\UI-MinusButton-Up:20:20|t"
			end
		else
			indentation = (header_main and "      " or "")..(header_child and "      " or "")
		end
		local repIs = barValue - barMin
		local repMax = barMax - barMin
		local color_atwar = atWarWith and "|cffff0000" or canToggleAtWar and "|cffffff00" or "|cff00ff00"
		local color_standing_blizzard = FACTION_COLORS_BLIZZARD[standingID]
		local color_standing_inc = FACTION_COLORS_INC[standingID]
		if not isHeader or hasRep then
			curLine = tooltip:AddLine()
			tooltip:SetCell(curLine, 1, indentation..(isHeader and "|cffffffff" or color_atwar)..name)
			tooltip:SetCell(curLine, 2, color_standing_blizzard..GetText("FACTION_STANDING_LABEL"..standingID, gender))
			tooltip:SetCell(curLine, 3, ("%s%5d / %5d"):format(color_standing_inc, repIs, repMax))
			tooltip:SetCell(curLine, 4, ("%s%.1f%%"):format(color_standing_inc, repIs / repMax * 100))
			local repDelta = reputationSessionBase[name] and (barValue - reputationSessionBase[name]) or 0
			tooltip:SetCell(curLine, 5, GetFactionDeltaColor(repDelta)..GetFactionDeltaText(repDelta))
		else
			curLine = tooltip:AddLine(indentation.."|cffffffff"..name)
		end
		tooltip:SetLineScript(curLine, "OnMouseUp", tooltip_line_OnMouseUp, factionIndex)
	end
	tooltip:UpdateScrolling()
end

function ElkFactions:GetFactionText(pattern, factionIndex, useColor)
	local name, description, standingID, barMin, barMax, barValue, atWarWith, canToggleAtWar, isHeader, isCollapsed, hasRep, isWatched, isChild = GetFactionInfo(factionIndex)
	local repIs = barValue - barMin
	local repMax = barMax - barMin
	local repDelta = reputationSessionBase[name] and (barValue - reputationSessionBase[name]) or 0
	
	pattern = "{cw}{n}{cr}: {cb}{s}{cr} - {ci}{rc}/{rm} ({rp}){cr} [{cd}{rd}{cr}]"
	
	--[[
		{n}    name
		{s}    standing
		{rc}   current reputation for this level
		{rm}   maximum reputation for this level
		{rp}   reputation percentage for this level
		{rd}   reputation change this session
		{cw}   color: at war
		{cb}   color: Blizzard
		{ci}   color: incremential
		{cd}   color: reputation change this session
		{cr}   color: reset
	]]
	local color_atwar = atWarWith and "|cffff0000" or canToggleAtWar and "|cffffff00" or "|cff00ff00"
	pattern = string_gsub(pattern, "%b{}", {
		["{n}"] = name,
		["{s}"] = GetText("FACTION_STANDING_LABEL"..standingID, gender),
		["{rc}"] = repIs,
		["{rm}"] = repMax,
		["{rp}"] = string_format("%.1f%%%%", repIs / repMax * 100),
		["{rd}"] = GetFactionDeltaText(repDelta),
		["{cw}"] = useColor and color_atwar or "",
		["{cb}"] = useColor and FACTION_COLORS_BLIZZARD[standingID] or "",
		["{ci}"] = useColor and FACTION_COLORS_INC[standingID] or "",
		["{cd}"] = GetFactionDeltaColor(repDelta),
		["{cr}"] = useColor and "|r" or "",
	})
	
	return pattern
end

local collapseQueue = {}
function ElkFactions:CollapseHeaders()
	local numFactions = GetNumFactions()
	local collapsedHeaders = self.db.char.collapsedHeaders
	local name, description, standingID, barMin, barMax, barValue, atWarWith, canToggleAtWar, isHeader, isCollapsed, hasRep, isWatched, isChild
	for factionIndex=1, numFactions do
		name, description, standingID, barMin, barMax, barValue, atWarWith, canToggleAtWar, isHeader, isCollapsed, hasRep, isWatched, isChild = GetFactionInfo(factionIndex)
		if isHeader and collapsedHeaders[name] and not isCollapsed then
			table.insert(collapseQueue, factionIndex)
		end
	end
	for factionIndex = #collapseQueue, 1, -1 do
		CollapseFactionHeader(collapseQueue[factionIndex])
	end
	table.wipe(collapseQueue)
end

function ElkFactions:GetAutoZoneFaction()
	local currentZone = GetRealZoneText()
	local currentSubZone = GetSubZoneText()
	
	return self.db.profile.AutoZoneFaction[currentZone.." - "..currentSubZone] or self.db.profile.AutoZoneFaction[currentZone]
end

function ElkFactions:ApplyAutoZoneFaction()
	if not self.db.profile.autoZoneFactionEnabled then return end
	
	local autoZoneFaction = self:GetAutoZoneFaction()
	-- is there a auto-faction set for the current zone?
	if autoZoneFaction then
		-- if lastFaction is set, last zone had autofaction, too
		-- if not we have to store the current watched faction
		if self.db.char.lastFaction == nil then
			self.db.char.lastFaction = GetWatchedFactionInfo() or false
		end
		self.SettingAutoFaction = true
		SetWatchedFactionName(autoZoneFaction)
	-- no auto-faction set so restore last faction if set
	elseif self.db.char.lastFaction ~= nil then
		self.SettingAutoFaction = true
		SetWatchedFactionName(self.db.char.lastFaction)
		self.db.char.lastFaction = nil
	end
end

function ElkFactions:Hook_SetWatchedFactionIndex(factionIndex)

--[===[@debug@
	self:Print("SetWatchedFactionIndex("..tostring(factionIndex)..") called...")
--@end-debug@]===]

	-- store index for lookup
	watchedFactionIndex = factionIndex and factionIndex > 0 and factionIndex or nil

	-- if we aren't setting the auto-faction
	if self.SettingAutoFaction then
		self.SettingAutoFaction = nil
	else
		self.db.char.lastFaction = factionIndex and factionIndex > 0 and GetFactionInfo(factionIndex) or nil
		self:ApplyAutoZoneFaction()
	end
end

function ElkFactions:Hook_CollapseFactionHeader(factionIndex)
	local name, description, standingID, barMin, barMax, barValue, atWarWith, canToggleAtWar, isHeader, isCollapsed, hasRep, isWatched, isChild = GetFactionInfo(factionIndex)

--[===[@debug@
	self:Print("CollapseFactionHeader("..tostring(factionIndex)..") called...")
	self:Print("    name: "..tostring(name), "isHeader: "..tostring(isHeader), "isCollapsed: "..tostring(isCollapsed))
--@end-debug@]===]

	if not (name and isHeader) then return end
	self.db.char.collapsedHeaders[name] = true
end

function ElkFactions:Hook_ExpandFactionHeader(factionIndex)
	local name, description, standingID, barMin, barMax, barValue, atWarWith, canToggleAtWar, isHeader, isCollapsed, hasRep, isWatched, isChild = GetFactionInfo(factionIndex)

--[===[@debug@
	self:Print("ExpandFactionHeader("..tostring(factionIndex)..") called...")
	self:Print("    name: "..tostring(name), "isHeader: "..tostring(isHeader), "isCollapsed: "..tostring(isCollapsed))
--@end-debug@]===]

	if not (name and isHeader) then return end
	self.db.char.collapsedHeaders[name] = nil
end

